/*
 * Licensed Materials - Property of IBM
 *
 * com.ibm.etools.c
 *
 *  Copyright IBM Corporation 2004. All Rights Reserved. 
 *
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.etools.c.importer;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.cdt.dstore.core.model.DataElement;

import com.ibm.etools.c.CClassifier;
import com.ibm.etools.c.CDerivableType;
import com.ibm.etools.c.CEnumerator;
import com.ibm.etools.c.CFactory;
import com.ibm.etools.c.CField;
import com.ibm.etools.c.CFunction;
import com.ibm.etools.c.CParameter;
import com.ibm.etools.c.CStruct;
import com.ibm.etools.c.CStructured;
import com.ibm.etools.c.CTypedef;
import com.ibm.etools.c.CUnion;
import com.ibm.etools.c.datatypes.CDirectionKind;
import com.ibm.etools.c.datatypes.CEnumeration;
import com.ibm.etools.c.datatypes.CVoid;
import com.ibm.etools.c.datatypes.DatatypesFactory;
import com.ibm.etools.typedescriptor.InstanceTDBase;

/**
 * CSourceUnit represents a source unit which can be a C source file or header
 * file. It has an associated object list which contains all of the type and
 * function objects within this source file. The importSource() method is the
 * exported interface.
 */
public class CSourceUnit {

    public static final String copyright = "Licensed Material - Property of IBM\n com.ibm.etools.c\n(C) Copyright IBM Corp. 2004 - All Rights Reserved.\nUS Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp."; //$NON-NLS-1$

    private CModelConverter _modelConverter;

    // The file name of the source unit
    private String _fileName;

    // The assocaited DataElement object
    private DataElement _source;

    // The list to keep all the C model objects in this source unit
    private ArrayList _objectList;

    // The list to keep all top level types.
    private ArrayList _topTypesList;

    // The type table from CModelConverter
    private CTypeTable _typeTable;

    // The type parser from CModelConverter
    private CTypeParser _typeParser;

    // The list of all include files
    private ArrayList _includeFiles;

    // A factory used to generate C language entities.
    private CFactory _factory;

    // A factory used to generate C datatypes.
    private DatatypesFactory _datatypesFactory;

    // Is this the first pass thru the file
    private boolean _firstPass;

    /**
     * Create a CSourceUnit object from a given source file name and a
     * CModelConverter object.
     */
    public CSourceUnit(String fileName, CModelConverter modelConverter) {
        _fileName = fileName;
        _modelConverter = modelConverter;
        _source = null;
        _objectList = new ArrayList();
        _topTypesList = new ArrayList();
        _typeTable = modelConverter.getTypeTable();
        _typeParser = modelConverter.getTypeParser();
        _includeFiles = new ArrayList();
        //_factory = modelConverter.getImporterResource().getFactory();
        //_datatypesFactory =
        // modelConverter.getImporterResource().getDatatypesFactory();
        _factory = CFactory.eINSTANCE;
        _datatypesFactory = DatatypesFactory.eINSTANCE;
        _firstPass = true;
    }

    /**
     * Create a CSourceUnit object from a given DataElement object which
     * represents the source file and a CModelConverter object.
     */
    public CSourceUnit(DataElement source, CModelConverter modelConverter) {
        _source = source;
        _fileName = source.getName();
        _modelConverter = modelConverter;
        _objectList = new ArrayList();
        _topTypesList = new ArrayList();
        _typeTable = modelConverter.getTypeTable();
        _typeParser = modelConverter.getTypeParser();
        _includeFiles = new ArrayList();
        //_factory = modelConverter.getImporterResource().getFactory();
        //_datatypesFactory =
        // modelConverter.getImporterResource().getDatatypesFactory();
        _factory = CFactory.eINSTANCE;
        _datatypesFactory = DatatypesFactory.eINSTANCE;
        _firstPass = true;
    }

    // getter methods
    public ArrayList getObjectList() {
        return _objectList;
    }

    public String getName() {
        return _fileName;
    }

    public DataElement getSource() {
        return _source;
    }

    public ArrayList getIncludeFiles() {
        return _includeFiles;
    }

    /**
     * Add the parsed content of a single source file to the MOF object list
     * This method is the entry point of the parsing process.
     */
    public ArrayList importSource() throws Exception {

        // Add all include files to the include list
        ArrayList includeElements = _source.getAssociated(CIM.P_INCLUDES_RELATION);

        for (int i = 0; i < includeElements.size(); i++) {
            DataElement includeElement = (DataElement) includeElements.get(i);
            String includeFileName = includeElement.getName();
            _includeFiles.add(includeFileName);
        }

        if (_source.getNestedSize() > 0) {

            for (int i = 0; i < _source.getNestedSize(); i++) {
                DataElement child = (DataElement) _source.get(i);
                addNode(child);
            }

            _firstPass = false;
            for (int i = 0; i < _source.getNestedSize(); i++) {
                DataElement child = (DataElement) _source.get(i);
                addNode(child);
            }

            // Set the offset and size for all struct/union fields.
            // We do this as the last step when all the type information is
            // known.
            if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                for (int i = 0; i < _objectList.size(); i++) {
                    Object object = _objectList.get(i);
                    if (object instanceof CStructured)
                        CTypeDescriptorBuilder.setInstancesOffsetAndSize((CStructured) object);
                }
            }
        }

        return _topTypesList;
    }

    /**
     * Add a node in the DataElement tree to the MOF object list
     */
    private void addNode(DataElement token) throws Exception {

        String type = token.getType();

        // Determine the node type and call the corresponding add mothod.
        if (type.equals(CIM.P_STRUCT_TYPE)) {
            addStruct(token, true);
        } else if (type.equals(CIM.P_UNION_TYPE)) {
            addUnion(token, true);
        } else if (type.equals(CIM.P_TYPEDEF_TYPE)) {
            addTypedef(token);
        } else if (type.equals(CIM.P_ENUMERATION_TYPE)) {
            addEnum(token);
        }
        // Don't parse functions - nothing done with results anyway.
        //	else if (type.equals(CIM.P_FUNCTION_TYPE) ||
        // type.equals(CIM.P_MAIN_FUNCTION_TYPE)) {
        //	 addFunction(token);
        //	}
        else if (CPreferenceStore.isIncludeGlobalVariables() && type.equals(CIM.P_VARIABLE_TYPE)) {
            addVariable(token);
        } else {
            // Ignore other tokens
        }
    }

    /**
     * Add a node representing a C structure to the object list
     */
    private CStruct addStruct(DataElement structToken, boolean topLevel) throws Exception {

        String structName = structToken.getName();
        CStruct astruct = (CStruct) _typeTable.getType("struct " + structName);
        if (astruct == null || astruct.getContains().size() == 0) {
            if (astruct == null) {
                astruct = _factory.createCStruct();
                if (structName.startsWith(CIM.P_UNNAMED_STRUCT)) {
                    astruct.setName("");
                } else {
                    astruct.setName(structName);
                }
                _typeTable.addType("struct " + structName, astruct);
            }

            _typeTable.addSourceLocation(astruct, new CSourceLocation(_fileName));

            for (int i = 0; i < structToken.getNestedSize(); i++) {
                DataElement child = (DataElement) structToken.get(i);
                String childType = child.getType();
                if (childType.equals(CIM.P_FIELD_TYPE)) {
                    CField afield = addField(child);
                    astruct.getContains().add(afield);
                    if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                        afield.setInstanceTDBase(CTypeDescriptorBuilder.createInstanceForField(afield));
                    }
                } else if (childType.equals(CIM.P_STRUCT_TYPE)) {

                    // Nested types also live in the global namespace in C.
                    // This will add the nested type to the type table and the
                    // object list.
                    CStruct nestedStruct = addStruct(child, false);

                    // If the nested struct contributes to the size of the outer
                    // struct,
                    // we will create a clone of it and add the cloned struct to
                    // the
                    // outer struct's containing list.
                    // The cloned struct has its own fields. The Type Descriptor
                    // instances of
                    // the fields may have different offsets than the original
                    // struct.
                    if (_typeParser.nestedTypeHasSizeEffect(child)) {
                        CStruct clonedStruct = clonedStruct(nestedStruct);
                        astruct.getContains().add(clonedStruct);
                    }
                } else if (childType.equals(CIM.P_UNION_TYPE)) {
                    CUnion nestedUnion = addUnion(child, false);
                    if (_typeParser.nestedTypeHasSizeEffect(child)) {
                        CUnion clonedUnion = clonedUnion(nestedUnion);
                        astruct.getContains().add(clonedUnion);
                    }
                } else if (childType.equals(CIM.P_ENUMERATION_TYPE)) {
                    CEnumeration nestedEnum = addEnum(child);
                }
            }

            if (topLevel)
                _topTypesList.add(astruct);

            _objectList.add(astruct);
        }
        return astruct;
    }

    /**
     * Add a node representing a C union to the object list
     */
    private CUnion addUnion(DataElement unionToken, boolean topLevel) throws Exception {

        String unionName = unionToken.getName();
        CUnion aunion = (CUnion) _typeTable.getType("union " + unionName);
        if (aunion == null || aunion.getContains().size() == 0) {
            if (aunion == null) {
                aunion = _factory.createCUnion();
                if (unionName.startsWith(CIM.P_UNNAMED_UNION)) {
                    aunion.setName("");
                } else {
                    aunion.setName(unionName);
                }
                _typeTable.addType("union " + unionName, aunion);
            }

            _typeTable.addSourceLocation(aunion, new CSourceLocation(_fileName));

            for (int i = 0; i < unionToken.getNestedSize(); i++) {
                DataElement child = (DataElement) unionToken.get(i);
                String childType = child.getType();
                if (childType.equals(CIM.P_FIELD_TYPE)) {
                    CField afield = addField(child);
                    aunion.getContains().add(afield);
                    if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                        afield.setInstanceTDBase(CTypeDescriptorBuilder.createInstanceForField(afield));
                    }
                } else if (childType.equals(CIM.P_STRUCT_TYPE)) {
                    CStruct nestedStruct = addStruct(child, false);
                    if (_typeParser.nestedTypeHasSizeEffect(child)) {
                        CStruct clonedStruct = clonedStruct(nestedStruct);
                        aunion.getContains().add(clonedStruct);
                    }
                } else if (childType.equals(CIM.P_UNION_TYPE)) {
                    CUnion nestedUnion = addUnion(child, false);
                    if (_typeParser.nestedTypeHasSizeEffect(child)) {
                        CUnion clonedUnion = clonedUnion(nestedUnion);
                        aunion.getContains().add(clonedUnion);
                    }
                } else if (childType.equals(CIM.P_ENUMERATION_TYPE)) {
                    CEnumeration nestedEnum = addEnum(child);
                }
            }

            if (topLevel)
                _topTypesList.add(aunion);

            _objectList.add(aunion);
        }

        return aunion;
    }

    /**
     * Add a node representing a C typedef to the object list
     */
    private CTypedef addTypedef(DataElement typedefToken) throws Exception {

        CTypedef atypedef = null;
        String typedefName = typedefToken.getName();
        CDerivableType atype = _typeTable.getType(typedefName);

        if (atype == null)
            atypedef = null;
        else if (atype instanceof CTypedef)
            atypedef = (CTypedef) atype;
        else
            return null;

        if (atypedef == null || atypedef.getDerives() == null) {
            if (atypedef == null) {
                atypedef = _factory.createCTypedef();
                atypedef.setName(typedefName);
                _typeTable.addType(typedefName, atypedef);
            }

            _typeTable.addSourceLocation(atypedef, new CSourceLocation(_fileName));
            // Add the deriving type
            CDerivableType derivingType = null;
            try {
                derivingType = _typeParser.findOrCreateType(typedefToken, true);
            } catch (Exception e) {
                // The first time thru ignore the errors, this might be a
                // forward reference.
                if (!_firstPass) {
                    if (CPreferenceStore.isThrowExceptionAtParseError())
                        throw e;
                    else
                        CResource.writeLog(this, "addTypedef", e);
                }
            }

            if (derivingType != null) {
                atypedef.setDerives(derivingType);
            } else if (!_firstPass)
                CResource.writeLog(this, "addTypedef", CResource.getString("_ERROR_Type_Not_Found", "typedef", typedefName));

            _objectList.add(atypedef);
            _topTypesList.add(atypedef);
        }
        return atypedef;
    }

    /**
     * Add an enum to the object list
     */
    private CEnumeration addEnum(DataElement enumToken) throws Exception {

        String enumName = enumToken.getName();

        CEnumeration anenum = (CEnumeration) _typeTable.getType("enum " + enumName);
        if (anenum == null || anenum.getEnumerator().size() == 0) {
            if (anenum == null) {
                anenum = _datatypesFactory.createCEnumeration();

                if (enumName.startsWith(CIM.P_UNNAMED_ENUM))
                    anenum.setName("");
                else
                    anenum.setName(enumName);

                _typeTable.addType("enum " + enumName, anenum);
            }

            _typeTable.addSourceLocation(anenum, new CSourceLocation(_fileName));
            for (int i = 0; i < enumToken.getNestedSize(); i++) {
                DataElement child = (DataElement) enumToken.get(i);
                if (child.getType().equals(CIM.P_ENUMERATOR_TYPE)) {
                    CEnumerator enumerator = addEnumerator(child);
                    anenum.getEnumerator().add(enumerator);
                }
            }
            _objectList.add(anenum);
            _topTypesList.add(anenum);
        }

        return anenum;
    }

    /**
     * Add a node representing a C function to the object list
     */
    private CFunction addFunction(DataElement functionToken) throws Exception {

        String functionName = functionToken.getName();
        CFunction afunction = _typeTable.getFunction(functionName);
        if (afunction == null) {
            afunction = _factory.createCFunction();
            afunction.setName(functionName);
            afunction.setIsVarArg(Boolean.FALSE);

            // Add parameters to the list
            ArrayList argumentList = functionToken.getAssociated(CIM.P_ARGUMENTS_RELATION);
            for (int i = 0; i < argumentList.size(); i++) {
                DataElement argument = (DataElement) argumentList.get(i);
                if (argument.getType().equals(CIM.P_TYPE_RELATION)) {
                    String argumentSignature = argument.getName();
                    if (argumentSignature != null && argumentSignature.length() > 0) {
                        // Set the IsVarArg flag if the last argument is "...".
                        if (argumentSignature.equals(CIM.P_FUNCTION_VARIABLE_ARGUMENT)) {
                            afunction.setIsVarArg(Boolean.TRUE);
                        } else {
                            CTypedElementParser atypedElement = new CTypedElementParser(argumentSignature);
                            CParameter aparameter = addParameter(atypedElement);
                            afunction.getParameter().add(aparameter);
                            // Create a TypeDescriptor instance for the
                            // parameter.
                            if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                                aparameter.setInstanceTDBase(CTypeDescriptorBuilder.createInstanceForVariable(aparameter));
                            }
                        }
                    }
                }
            }

            // Add the return value to the list
            ArrayList returnList = functionToken.getAssociated(CIM.P_RETURN_TYPE_RELATION);

            CParameter returnParam = addReturn(returnList);

            if (returnParam != null) {
                afunction.getParameter().add(returnParam);
                if (CPreferenceStore.isPopulateTypeDescriptorModel() && !(returnParam.getType() instanceof CVoid))
                    returnParam.setInstanceTDBase(CTypeDescriptorBuilder.createInstanceForVariable(returnParam));
            }

            _typeTable.addFunction(functionName, afunction);
            _typeTable.addSourceLocation(afunction, new CSourceLocation(_fileName));
            _objectList.add(afunction);
            _topTypesList.add(afunction);
        }
        return afunction;
    }

    /**
     * Add a node representing a field of a structure/union to the object list
     */
    private CField addField(DataElement fieldToken) throws Exception {

        CField afield = _factory.createCField();
        String fieldName = fieldToken.getName();
        if (fieldName.startsWith(CIM.P_UNNAMED_BITFIELD))
            afield.setName("");
        else
            afield.setName(fieldName);

        // Get the type of the field
        CDerivableType fieldType = null;
        try {
            fieldType = _typeParser.findOrCreateType(fieldToken, false);
        } catch (Exception e) {
            if (CPreferenceStore.isThrowExceptionAtParseError())
                throw e;
            else
                CResource.writeLog(this, "addField", e);
        }

        if (fieldType instanceof CClassifier) {
            afield.setType((CClassifier) fieldType);
        } else
            CResource.writeLog(this, "addField", CResource.getString("_ERROR_Type_Not_Found", "field", fieldName));

        return afield;
    }

    /**
     * Add a global variable to the object list
     */
    private CField addVariable(DataElement varToken) throws Exception {

        CField avar = addField(varToken);
        if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
            avar.setInstanceTDBase(CTypeDescriptorBuilder.createInstanceForVariable(avar));
        }
        _objectList.add(avar);
        _topTypesList.add(avar);
        return avar;
    }

    /**
     * Add a node representing an enumerator to the object list
     */
    private CEnumerator addEnumerator(DataElement enumeratorToken) throws Exception {

        CEnumerator enumerator = _factory.createCEnumerator();

        String enumeratorString = enumeratorToken.getValue();
        String enumeratorName = enumeratorString;
        int index = enumeratorString.indexOf('=');
        if (index > 0) {
            enumeratorName = enumeratorString.substring(0, index).trim();
            String valueString = enumeratorString.substring(index + 1).trim();

            Integer value = new Integer(0);
            try {
                value = Integer.valueOf(valueString);
            } catch (Exception e) {
                if (CPreferenceStore.isThrowExceptionAtParseError())
                    throw e;
                else
                    CResource.writeLog(this, "addEnumerator", e);
            }

            enumerator.setValue(value);
        }

        enumerator.setName(enumeratorName);
        return enumerator;
    }

    /**
     * Add a node representing a function parameter to the object list
     */
    private CParameter addParameter(DataElement paramToken) throws Exception {

        CParameter aparameter = _factory.createCParameter();

        // **TO FIX
        //aparameter.setDirection(CIM.M_DIRECTION_ARGUMENT);
        aparameter.setDirection(CDirectionKind.PARAMETER_LITERAL);
        aparameter.setName(paramToken.getName());

        // Get the parameter type
        CDerivableType paramType = null;
        try {
            paramType = _typeParser.findOrCreateType(paramToken, true);
        } catch (Exception e) {
            if (CPreferenceStore.isThrowExceptionAtParseError())
                throw e;
            else
                CResource.writeLog(this, "addParameter", e);
        }

        if (paramType instanceof CClassifier)
            aparameter.setType((CClassifier) paramType);
        else
            CResource.writeLog(this, "addParameter", CResource.getString("_ERROR_Type_Not_Found", "parameter", paramToken.getName()));

        return aparameter;
    }

    /**
     * Add a node representing a function parameter to the object list
     */
    private CParameter addParameter(CTypedElementParser atypedElement) throws Exception {

        CParameter aparameter = _factory.createCParameter();
        // **TO FIX
        //aparameter.setDirection(CIM.M_DIRECTION_ARGUMENT);
        aparameter.setDirection(CDirectionKind.PARAMETER_LITERAL);
        aparameter.setName(atypedElement.getName());

        // Get the parameter type
        CDerivableType paramType = null;
        try {
            paramType = _typeParser.findOrCreateType(atypedElement.getType(), true);
        } catch (Exception e) {
            if (CPreferenceStore.isThrowExceptionAtParseError())
                throw e;
            else
                CResource.writeLog(this, "addParameter", e);
        }

        if (paramType instanceof CClassifier)
            aparameter.setType((CClassifier) paramType);
        else
            CResource.writeLog(this, "addParameter", CResource.getString("_ERROR_Type_Not_Found", "parameter", atypedElement.getName()));

        return aparameter;
    }

    /**
     * Add the return parameter of a function to the list
     */
    private CParameter addReturn(ArrayList returnList) throws Exception {

        if (returnList == null || returnList.size() == 0)
            return null;

        CParameter areturn = _factory.createCParameter();
        //**TO FIX
        //areturn.setDirection(CIM.M_DIRECTION_RETURN);
        areturn.setDirection(CDirectionKind.RETURN_LITERAL);
        areturn.setName("");

        String returnTypeString = _typeParser.typeListToString(returnList);

        CDerivableType returnType = null;
        try {
            returnType = _typeParser.findOrCreateType(returnTypeString, true);
        } catch (Exception e) {
            if (CPreferenceStore.isThrowExceptionAtParseError())
                throw e;
            else
                CResource.writeLog(this, "addReturn", e);
        }

        if (returnType instanceof CClassifier)
            areturn.setType((CClassifier) returnType);
        else
            CResource.writeLog(this, "addReturn", CResource.getString("_ERROR_Type_Not_Found", "return parameter", returnTypeString));

        return areturn;
    }

    /**
     * Return a clone of the given CStruct object. This is used in the case of
     * nested structs.
     */
    private CStruct clonedStruct(CStruct astruct) {

        CStruct clonedStruct = _factory.createCStruct();
        clonedStruct.setName(astruct.getName());
        List fields = astruct.getContains();
        Iterator it = fields.iterator();
        while (it.hasNext()) {
            Object nextObject = it.next();
            if (nextObject instanceof CField) {
                CField afield = (CField) nextObject;
                CField clonedField = _factory.createCField();
                clonedField.setName(afield.getName());
                clonedField.setType(CQuery.typeOf(afield));
                if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                    InstanceTDBase clonedInstance = CTypeDescriptorBuilder.clonedInstance(afield.getInstanceTDBase());
                    clonedField.setInstanceTDBase(clonedInstance);
                }
                clonedStruct.getContains().add(clonedField);
            } else if (nextObject instanceof CStruct) {
                CStruct nestedStruct = clonedStruct((CStruct) nextObject);
                clonedStruct.getContains().add(nestedStruct);
            } else if (nextObject instanceof CUnion) {
                CUnion nestedUnion = clonedUnion((CUnion) nextObject);
                clonedStruct.getContains().add(nestedUnion);
            }
        }
        return clonedStruct;
    }

    /**
     * Return a clone of the given CUnion object. This is used in the case of
     * nested unions.
     */
    private CUnion clonedUnion(CUnion aunion) {

        CUnion clonedUnion = _factory.createCUnion();
        clonedUnion.setName(aunion.getName());
        List fields = aunion.getContains();
        Iterator it = fields.iterator();
        while (it.hasNext()) {
            Object nextObject = it.next();
            if (nextObject instanceof CField) {
                CField afield = (CField) nextObject;
                CField clonedField = _factory.createCField();
                clonedField.setName(afield.getName());
                clonedField.setType(CQuery.typeOf(afield));
                if (CPreferenceStore.isPopulateTypeDescriptorModel()) {
                    InstanceTDBase clonedInstance = CTypeDescriptorBuilder.clonedInstance(afield.getInstanceTDBase());
                    clonedField.setInstanceTDBase(clonedInstance);
                }
                clonedUnion.getContains().add(clonedField);
            } else if (nextObject instanceof CStruct) {
                CStruct nestedStruct = clonedStruct((CStruct) nextObject);
                clonedUnion.getContains().add(nestedStruct);
            } else if (nextObject instanceof CUnion) {
                CUnion nestedUnion = clonedUnion((CUnion) nextObject);
                clonedUnion.getContains().add(nestedUnion);
            }
        }
        return clonedUnion;
    }

}